@@ -2,28 +2,34 @@
 #include <stdbool.h>
 
 #include "backend/backend_common.h"
+#include "utils/misc.h"
 
 #include "gl_common.h"
 
 struct gl_blur_context {
-	enum blur_method method;
 	struct gl_shader *blur_shader;
 
 	/// Temporary textures used for blurring
 	GLuint *blur_textures;
-	int blur_texture_count;
 	/// Temporary fbos used for blurring
 	GLuint *blur_fbos;
+
+	int blur_texture_count;
 	int blur_fbo_count;
 
-	/// Cached dimensions of each blur_texture. They are the same size as the target,
-	/// so they are always big enough without resizing.
+	/// Cached dimensions of each blur_texture. They keep growing to accommodate each
+	/// source image it gets. The hope is soon it gets big enough and we stop
+	/// resizing.
 	/// Turns out calling glTexImage to resize is expensive, so we avoid that.
 	struct texture_size {
 		int width;
 		int height;
 	} *texture_sizes;
 
+	/// Final render target size, used to detect root size change and triggers
+	/// blur_texture shrinking to reduce memory usage.
+	uint64_t root_size;
+
 	/// Cached dimensions of the offscreen framebuffer. It's the same size as the
 	/// target but is expanded in either direction by resize_width / resize_height.
 	int fb_width, fb_height;
@@ -34,6 +40,8 @@ struct gl_blur_context {
 	int npasses;
 
 	enum backend_image_format format;
+
+	enum blur_method method;
 };
 
 // TODO(yshui) small optimization for kernel blur, if source and target are different,
@@ -43,10 +51,11 @@ struct gl_blur_context {
 /**
  * Blur contents in a particular region.
  */
-static bool gl_kernel_blur(double opacity, struct gl_blur_context *bctx,
-                           const struct backend_mask_image *mask, const GLuint vao[2],
-                           const int vao_nelems[2], struct gl_texture *source,
-                           GLuint blur_sampler, GLuint target_fbo, GLuint default_mask) {
+static bool
+gl_kernel_blur(double opacity, struct gl_blur_context *bctx,
+               const struct backend_mask_image *mask, vec2 mask_scale,
+               const GLuint vao[2], const int vao_nelems[2], struct gl_texture *source,
+               GLuint blur_sampler, GLuint target_fbo, GLuint default_mask) {
 	int curr = 0;
 	for (int i = 0; i < bctx->npasses; ++i) {
 		auto p = &bctx->blur_shader[i];
@@ -87,6 +96,7 @@ static bool gl_kernel_blur(double opacity, struct gl_blur_context *bctx,
 		glUniform2f(UNIFORM_MASK_OFFSET_LOC, 0.0F, 0.0F);
 		glUniform1i(UNIFORM_MASK_INVERTED_LOC, 0);
 		glUniform1f(UNIFORM_MASK_CORNER_RADIUS_LOC, 0.0F);
+		glUniform2f(UNIFORM_MASK_SCALE_LOC, 1.0F, 1.0F);
 
 		// The number of indices in the selected vertex array
 		GLsizei nelems;
@@ -119,14 +129,17 @@ static bool gl_kernel_blur(double opacity, struct gl_blur_context *bctx,
 				glUniform1i(UNIFORM_MASK_INVERTED_LOC, mask->inverted);
 				glUniform1f(UNIFORM_MASK_CORNER_RADIUS_LOC,
 				            (float)mask->corner_radius);
-				glUniform2f(UNIFORM_MASK_OFFSET_LOC, (float)(mask->origin.x),
-				            (float)(mask->origin.y));
+				glUniform2f(UNIFORM_MASK_OFFSET_LOC,
+				            (float)mask->origin.x, (float)mask->origin.y);
+				glUniform2f(UNIFORM_MASK_SCALE_LOC, (float)mask_scale.x,
+				            (float)mask_scale.y);
 			}
 			glBindVertexArray(vao[0]);
 			nelems = vao_nelems[0];
 			glBindFramebuffer(GL_FRAMEBUFFER, target_fbo);
 
 			glUniform1f(UNIFORM_OPACITY_LOC, (float)opacity);
+			glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
 		}
 
 		glDrawElements(GL_TRIANGLES, nelems, GL_UNSIGNED_INT, NULL);
@@ -146,8 +159,8 @@ static bool gl_kernel_blur(double opacity, struct gl_blur_context *bctx,
 ///            [0]: for sampling from blurred result into the target fbo.
 ///            [1]: for sampling from the source texture into blurred textures.
 bool gl_dual_kawase_blur(double opacity, struct gl_blur_context *bctx,
-                         const struct backend_mask_image *mask, const GLuint vao[2],
-                         const int vao_nelems[2], struct gl_texture *source,
+                         const struct backend_mask_image *mask, vec2 mask_scale,
+                         const GLuint vao[2], const int vao_nelems[2], struct gl_texture *source,
                          GLuint blur_sampler, GLuint target_fbo, GLuint default_mask) {
 	int iterations = bctx->blur_texture_count;
 	int scale_factor = 1;
@@ -209,6 +222,7 @@ bool gl_dual_kawase_blur(double opacity, struct gl_blur_context *bctx,
 	glUniform2f(UNIFORM_MASK_OFFSET_LOC, 0.0F, 0.0F);
 	glUniform1i(UNIFORM_MASK_INVERTED_LOC, 0);
 	glUniform1f(UNIFORM_MASK_CORNER_RADIUS_LOC, 0.0F);
+	glUniform2f(UNIFORM_MASK_SCALE_LOC, 1.0F, 1.0F);
 	glUniform1f(UNIFORM_OPACITY_LOC, 1.0F);
 
 	for (int i = iterations - 1; i >= 0; --i) {
@@ -245,12 +259,15 @@ bool gl_dual_kawase_blur(double opacity, struct gl_blur_context *bctx,
 				            (float)mask->corner_radius);
 				glUniform2f(UNIFORM_MASK_OFFSET_LOC, (float)(mask->origin.x),
 				            (float)(mask->origin.y));
+				glUniform2f(UNIFORM_MASK_SCALE_LOC, (float)mask_scale.x,
+				            (float)mask_scale.y);
 			}
 			glBindVertexArray(vao[0]);
 			nelems = vao_nelems[0];
 			glBindFramebuffer(GL_DRAW_FRAMEBUFFER, target_fbo);
 
 			glUniform1f(UNIFORM_OPACITY_LOC, (GLfloat)opacity);
+			glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
 		}
 
 		glUniform1f(UNIFORM_SCALE_LOC, (GLfloat)scale_factor);
@@ -263,13 +280,19 @@ bool gl_dual_kawase_blur(double opacity, struct gl_blur_context *bctx,
 	return true;
 }
 
-static bool
-gl_blur_context_preallocate_textures(struct gl_blur_context *bctx, ivec2 source_size) {
-	if (source_size.width != bctx->fb_width || source_size.height != bctx->fb_height) {
-		// Resize the temporary textures used for blur in case the root
-		// size changed
-		bctx->fb_width = source_size.width;
-		bctx->fb_height = source_size.height;
+static bool gl_blur_context_preallocate_textures(struct gl_blur_context *bctx,
+                                                 ivec2 source_size, uint64_t root_size) {
+	if (root_size != bctx->root_size) {
+		bctx->root_size = root_size;
+		bctx->fb_width = 0;
+		bctx->fb_height = 0;
+	}
+
+	if (source_size.width > bctx->fb_width || source_size.height > bctx->fb_height) {
+		// Resize the temporary textures used for blur in case the source image
+		// can't fit.
+		bctx->fb_width = max2(bctx->fb_width, source_size.width);
+		bctx->fb_height = max2(bctx->fb_height, source_size.height);
 
 		for (int i = 0; i < bctx->blur_texture_count; ++i) {
 			auto tex_size = bctx->texture_sizes + i;
@@ -337,8 +360,9 @@ bool gl_blur(struct backend_base *base, ivec2 origin, image_handle target_,
 		return true;
 	}
 
+	uint64_t root_size = (uint64_t)gd->back_image.height * (uint64_t)gd->back_image.width;
 	if (!gl_blur_context_preallocate_textures(
-	        bctx, (ivec2){source->width, source->height})) {
+	        bctx, (ivec2){source->width, source->height}, root_size)) {
 		return false;
 	}
 
@@ -360,7 +384,7 @@ bool gl_blur(struct backend_base *base, ivec2 origin, image_handle target_,
 	// we never actually use that capability anywhere.
 	assert(source->y_inverted);
 
-	glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
+	glBlendFunc(GL_ONE, GL_ZERO);
 	glBindVertexArray(gd->vertex_array_objects[0]);
 	glBindBuffer(GL_ARRAY_BUFFER, gd->buffer_objects[0]);
 	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, gd->buffer_objects[1]);
@@ -391,15 +415,15 @@ bool gl_blur(struct backend_base *base, ivec2 origin, image_handle target_,
 
 	auto target_fbo = gl_bind_image_to_fbo(gd, (image_handle)target);
 	if (bctx->method == BLUR_METHOD_DUAL_KAWASE) {
-		ret = gl_dual_kawase_blur(args->opacity, bctx, args->source_mask,
-		                          gd->vertex_array_objects, vao_nelems, source,
-		                          gd->samplers[GL_SAMPLER_BLUR], target_fbo,
-		                          gd->default_mask_texture);
+		ret = gl_dual_kawase_blur(
+		    args->opacity, bctx, args->source_mask, args->source_mask_scale,
+		    gd->vertex_array_objects, vao_nelems, source,
+		    gd->samplers[GL_SAMPLER_BLUR], target_fbo, gd->default_mask_texture);
 	} else {
 		ret = gl_kernel_blur(args->opacity, bctx, args->source_mask,
-		                     gd->vertex_array_objects, vao_nelems, source,
-		                     gd->samplers[GL_SAMPLER_BLUR], target_fbo,
-		                     gd->default_mask_texture);
+		                     args->source_mask_scale, gd->vertex_array_objects,
+		                     vao_nelems, source, gd->samplers[GL_SAMPLER_BLUR],
+		                     target_fbo, gd->default_mask_texture);
 	}
 
 	glBindFramebuffer(GL_FRAMEBUFFER, 0);
@@ -617,7 +641,7 @@ bool gl_create_kernel_blur_context(void *blur_context, GLfloat *projection,
 		// Build program
 		pass->prog = gl_create_program_from_strv(
 		    (const char *[]){vertex_shader, NULL},
-		    (const char *[]){shader_str, masking_glsl, NULL});
+		    (const char *[]){shader_str, scaled_masking_glsl, NULL});
 		free(shader_str);
 		if (!pass->prog) {
 			log_error("Failed to create GLSL program.");
@@ -642,7 +666,7 @@ bool gl_create_kernel_blur_context(void *blur_context, GLfloat *projection,
 		auto pass = &ctx->blur_shader[1];
 		pass->prog = gl_create_program_from_strv(
 		    (const char *[]){vertex_shader, NULL},
-		    (const char *[]){blend_with_mask_frag, masking_glsl, NULL});
+		    (const char *[]){blend_with_mask_frag, scaled_masking_glsl, NULL});
 
 		// Setup projection matrix
 		glUseProgram(pass->prog);
@@ -791,7 +815,7 @@ bool gl_create_dual_kawase_blur_context(void *blur_context, GLfloat *projection,
 		// Build program
 		up_pass->prog = gl_create_program_from_strv(
 		    (const char *[]){vertex_shader, NULL},
-		    (const char *[]){shader_str, masking_glsl, NULL});
+		    (const char *[]){shader_str, scaled_masking_glsl, NULL});
 		free(shader_str);
 		if (!up_pass->prog) {
 			log_error("Failed to create GLSL program.");
